"""
Bot sources module. Defines classes of distinct bot sources that behave
differently at the various steps of a factory functionality

Constants

    SUPPORTED_SOURCES:  Dictionary of <source_identifier>: <source_subclass>
"""
import os
import shutil
import subprocess
import hashlib
import sys

import git
from click import BadParameter
from colorama import Fore

from pyqalx.core.errors import QalxFactoryValidationError, QalxFactoryBuildError
from pyqalx.core.utils import _msg_user


class BotSource:
    """
    Base class for source related functionality about a specific bot

    Instance attributes

        bot_name:       The bot name
        bot:            The entire bot definition from the plan
        build_path:     The bot build path
        md5_hash:       The md5_hash of the bot code
    """

    FORE = Fore.MAGENTA

    def __init__(self, bot_name, bot, build_path_dir, log, *args, **kwargs):
        self.bot_name = bot_name
        self.bot = bot
        self.source_dict = bot["source"]
        self.build_path = os.path.join(build_path_dir, bot_name)
        self.md5_hash = None
        self.log = log

    @staticmethod
    def get_source(bot):
        """
        Static method that accepts a bot definition and tries to identify
        the source type and return the appropriate BotSource subclass if found

        :param bot: The bot definition from the plan
        :return: The BotSource subclass if found

        :raises KeyError:     If no bot source subclass is identified
        """
        source = bot["source"]
        for key in source:
            bot_source_class = SUPPORTED_SOURCES.get(key.lower())
            if bot_source_class:
                return bot_source_class
        raise KeyError(f"Unknown bot source `{source}`")

    def _msg_user(self, msg, **kwargs):
        """
        Helper method that prints and also logs a debug message if a log
        instance attribute exists
        """
        _msg_user(
            title=self.bot_name, fore=self.FORE, msg=msg, log=self.log, **kwargs
        )

    def zip_build_path(self):
        """
        Zips the contents of the build_path. The name of the file is:
        "<name_of_build_path>.zip" and it is located in the parent folder of
        build_path

        :return: The path to the zip file created
        """
        return shutil.make_archive(self.build_path, "zip", self.build_path)

    @staticmethod
    def get_single_hash(file_path):
        """
        The an md5 hash for a single file. Bytes are read in chunks for a given
        file

        :param file_path: The path to the file to get an md5 hash for
        :return: The md5 hash
        """
        hash_md5 = hashlib.md5()
        with open(file_path, "rb") as f:
            # Reading the file in chunks to avoid reading the whole file in
            # memory: https://stackoverflow.com/a/3431838
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def get_hash(self):
        """
        Calculates the md5 hash of the contents of the build path, the value of
        which is stored under the md5_hash instance attribute
        """
        self.md5_hash = self._get_hash()

    def _get_hash(self):
        """
        Hook for calculating the md5 hash for each bot source type
        """
        self._msg_user("Calculating hash")
        hashes = []
        for root, _, files in os.walk(self.build_path):
            for file in files:
                fp = os.path.join(root, file)
                hashes.append(self.get_single_hash(fp))
        hasher = hashlib.md5()
        for hash_value in sorted(hashes):
            hasher.update(hash_value.encode("utf-8"))
        self._msg_user("Finished calculating hash")
        return hasher.hexdigest()

    def validate(self):
        """
        Hook for providing validation on a BotSource. Called by `download_bot`
        after the download has completed
        """
        self._validate_bot_path()

    def _validate_bot_path(self):
        """
        Validates that the bot_path is importable.  Adds the build_path to the
        path temporarily so that we can attempt the import.
        """
        from qalxcli.cli_types import BOT_IMPORT

        sys.path.append(self.build_path)

        def _remove_from_path():
            sys.path.remove(self.build_path)

        try:
            BOT_IMPORT(self.bot["bot_path"])
        except BadParameter as exc:
            _remove_from_path()
            raise QalxFactoryValidationError(
                f"Error validating `bot_path` on `{self.bot_name}`: {exc}"
            )
        _remove_from_path()

    def _download_bot(self):
        """Hook for downloading a bot for a specific BotSource"""
        pass

    def download_bot(self):
        bot = self._download_bot()
        self.validate()
        return bot

    def install_bot(self):
        """
        Install the bot code. This is needed for the bot code package and the
        dependencies to be correctly installed and configured
        """
        self._msg_user("Installing...")
        self._install_bot()
        self._msg_user("Finished Installing...")

    def _install_bot(self):
        """
        Hook for installing the bot code using pip and code that has been
        downloaded in the build path
        """
        # First try to install from `pyproject.toml` or `setup.py`.
        output = subprocess.run(
            [shutil.which("pip"), "install", "."],
            cwd=self.build_path,
            check=False,
        )
        if output.returncode == 0:
            return
        self._msg_user(
            "Unable to install dependencies from `setup.py` "
            "or `pyproject.toml`.  "
            "Attempting to install from requirements file."
        )

        # The fallback option is to use the user defined requirements.txt file
        # if specified - failing that attempt to use a `requirements.txt` file
        # on `self.bot_path`
        requirements_path_key = "requirements_path"
        requirements_path = self.bot.get(
            requirements_path_key, "requirements.txt"
        )
        if requirements_path_key in self.bot and not os.path.exists(
            requirements_path
        ):
            # If the user has specified requirements we should check the
            # file exists
            msg = f"{requirements_path} does not exist."
            raise QalxFactoryBuildError(msg)
        output = subprocess.run(
            [shutil.which("pip"), "install", "-r", requirements_path],
            cwd=self.build_path,
            check=False,
        )

        if output.returncode == 0:
            return
        if os.path.exists(requirements_path):
            # The requirements path existed but there was an error installing.
            # An exception is raised so the user can go and fix it
            raise QalxFactoryBuildError(
                f"An error occurred installing"
                f" dependencies "
                f"from {requirements_path}."
            )
        # There was no file found to install dependencies from.  This may be
        # intentional but we have no way of knowing.  Inform the user.
        self._msg_user(
            "No files found to install dependencies from."
            "  Dependencies not installed."
        )


class PathBotSource(BotSource):
    """
    Bot source class for code provide from a local path

    Instance attributes
        path:   The local path to the bot code
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = self.source_dict.get("path")

    def _download_bot(self):
        """
        Copies the local bot code to build_path
        """
        msg = f"Copying contents from `{self.path}` to `{self.build_path}`"
        self._msg_user(msg)
        shutil.copytree(self.path, self.build_path)


class GitBotSource(BotSource):
    """
    Bot source class for code provided from a url of a git repository

    Instance attributes
        url:   The url to the git repository with the bot code
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = self.source_dict.pop("url")

    def _download_bot(self):
        """
        Downloads the bot code to build_path
        """
        msg = f"Cloning repository from `{self.url}` to `{self.build_path}`"
        self._msg_user(msg)
        git.Repo.clone_from(self.url, self.build_path, **self.source_dict)


class PyPIBotSource(BotSource):
    """
    Bot source class for code provided from a pypi repository

    Class attributes

        PYPI:   The key for pypi on the bot source
    """

    PYPI = "pypi"

    @staticmethod
    def install(source_dict):
        """
        method that allows easier installation of pypi sources for remote
        sectors
        :param source_dict:The source dict
        """
        pypi = source_dict.get(PyPIBotSource.PYPI)
        version = source_dict.get("version")
        if version is not None:
            pypi += f"=={version}"

        output = subprocess.run(
            [shutil.which("pip"), "install", pypi], check=False
        )
        if output.returncode != 0:
            raise QalxFactoryBuildError(f"Failed to install {pypi}")

    def zip_build_path(self):
        """
        pypi sources don't get downloaded - so there is nothing to zip
        """

    def _install_bot(self):
        """
        Pypi bots get installed directly from source
        """
        PyPIBotSource.install(self.source_dict)

    def _validate_bot_path(self):
        """
        bot_path cannot be validated for pypi as the package won't
        be installed so there is nothing that can be imported.
        """

    def _download_bot(self):
        """
        `pypi` sources don't get downloaded.  This is because, if the associated
        bot is for a remote sector the OS could be different - meaning that
        what is downloaded may not work on the remote sector.  Instead, these
        are just installed when `build` is called. An item will still be created
        but it won't have any data on it.
        """
        self._msg_user("Found pypi source.  Not downloading")

    def _get_hash(self):
        """
        Because no file is downloaded there is nothing to calculate a hash on.
        The API cannot be queried as the hashes are calculated for a specific
        file type (.tar.gz, .whl etc) so there is no way to be consistent.
        """
        self._msg_user("Found pypi source.  Not generating hash")


class QalxItemBotSource(BotSource):
    """
    Source class for Items that contain Bot code as a zip file.  Used by
    remote sectors to download and extract the bot code.

    Instance attributes

        bot_name:       The bot name
        bot:            The entire bot definition from the plan
        build_path:     The bot build path
        bot_code_item:  The ~pyqalx.Item instance that contains the zipped bot
                        code
    """

    def __init__(self, bot_code_item, *args, **kwargs):
        super(QalxItemBotSource, self).__init__(*args, **kwargs)
        self.bot_code_item = bot_code_item

    def _install_bot(self):
        if self.source_dict.get(PyPIBotSource.PYPI):
            # ensure pypi bots get installed in a specific way
            PyPIBotSource.install(source_dict=self.source_dict)
        else:
            super(QalxItemBotSource, self)._install_bot()

    def _download_bot(self):
        """
        Downloads the bot code from the Item and saves it in the appropriate
        build path for this bot.  Unzips the zip file.  Installs the bot
        """
        msg = f"Downloading bot code from API to `{self.build_path}`"
        self._msg_user(msg)
        if not os.path.exists(self.build_path):
            # The <bot_name> sub directory probably doesn't exist at bootstrap
            # time.  Ensure it gets created
            os.mkdir(self.build_path)

        if not self.source_dict.get(PyPIBotSource.PYPI):
            # Save the zipfile to disk.  Pypi bots won't have a file
            filename = self.bot_code_item.save_file_to_disk(self.build_path)
            # Unzip the zipped bot code into the build path.  This path will get
            # added to the PYTHONPATH at bot start time.
            shutil.unpack_archive(
                filename=filename, extract_dir=self.build_path
            )
        # Bots on remote instances should always be installed into the
        # environment immediately after downloading
        self.install_bot()
        # Return the bot_name as this is required for starting the bot
        return self.bot_name


SUPPORTED_SOURCES = {
    "url": GitBotSource,
    "path": PathBotSource,
    "pypi": PyPIBotSource,
}



from pyqalx.core.adapters.adapter import (
    QalxNamedEntityAdapter,
    QalxUpdateStatusAdapter,
)
from pyqalx.core.entities import Bot


class QalxBot(QalxNamedEntityAdapter, QalxUpdateStatusAdapter):
    _entity_class = Bot
    _user_only_methods = ["add"]
    _bot_only_methods = ["replace_workers"]

    def add(self, name, config, meta=None, **kwargs):
        """
        Creates a `Bot` instance.

        :param name: The name that this bot will be given
        :type name: str
        :param config: The bots config
        :type config: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :return: The newly created `Bot` instance
        """
        return super(QalxBot, self).add(
            host=self.session._host_info,
            name=name,
            meta=meta,
            config=config,
            **kwargs,
        )

    def replace_workers(self, bot_entity, workers):
        """
        Completely replaces any Workers on the given bot.  Will return the
        replaced workers in an unpacked state

        :param bot_entity: The ~entities.bot.Bot entity that is being changed
        :param workers: The number of workers that this bot should have
        :return: A ~entities.bot.Bot instance with the updated workers
        """
        guid = bot_entity["guid"]
        detail_endpoint = self.detail_endpoint(guid=guid)
        endpoint = f"{detail_endpoint}/replace-workers"
        self.session.log.debug(
            f"replace workers {self.entity_class.entity_type} with"
            f" guid {guid} with {endpoint}"
        )

        resp = self._process_api_request("patch", endpoint, workers=workers)
        return self._process_api_response(resp)

    def _signal_workers(self, entity, signal_method, *args, **kwargs):
        """
        Helper method for calling signals on workers.  A bot doesn't have
        the concept of `signals`.  Instead, we call a signal method we just
        apply this to all the workers on the bot
        """

        for worker in entity.get("workers", []):
            worker_method = getattr(self.session.worker, signal_method)
            worker_method(worker, bot_entity=entity, *args, **kwargs)

    def stop(self, entity, **kwargs):
        """
        Stops all the workers on the bot
        """
        self._signal_workers(entity, signal_method="stop", **kwargs)

    def resume(self, entity, **kwargs):
        """
        Resumes all the workers on the bot
        """
        self._signal_workers(entity, signal_method="resume", **kwargs)

    def terminate(self, entity, *args, **kwargs):
        """
        Terminates all the workers on the bot
        """
        self._signal_workers(entity, signal_method="terminate", *args, **kwargs)



import os
from itertools import zip_longest

import requests

from pyqalx.core.encryption import QalxEncrypt
from pyqalx.core.entities.object_dict import ObjectDict

from pyqalx.core.errors import (
    QalxAPIResponseError,
    QalxEntityTypeNotFound,
    QalxError,
    QalxQueueError,
)


class QalxListEntity(ObjectDict):
    """
    Simple wrapper around a pyqalxapi_dict so we can keep extra keys
    on the API list response.  Instantiates each entity in `data` to the
    correct QalxEntity subclass.
    """

    _data_key = "data"

    def __new__(cls, pyqalxapi_list_response_dict, *args, **kwargs):
        """
        A QalxListEntity is just an ObjectDict that has a list of
        QalxEntity instances stored on the `_data_key`.
        :param pyqalxapi_list_response_dict: A dict that gets returned from
        a pyqalxapi list endpoint.  This should at minimum have a `data`
        key but may have other keys which we preserve
        :param kwargs: Must contain `child` key which is a subclass of `QalxEntity`

        """
        cls.child = kwargs["child"]

        if not issubclass(cls.child, (QalxEntity,)):
            raise QalxEntityTypeNotFound(
                f"Expected `child` to be a subclass of "
                f"`QalxEntity`.  Got `{cls.child}`"
            )

        return super(QalxListEntity, cls).__new__(
            cls, pyqalxapi_list_response_dict
        )

    def __init__(self, pyqalxapi_list_response_dict, *args, **kwargs):
        super().__init__(pyqalxapi_list_response_dict)

        if (
            self._data_key not in pyqalxapi_list_response_dict
            or not isinstance(
                pyqalxapi_list_response_dict[self._data_key], list
            )
        ):
            raise QalxAPIResponseError(
                "Expected `{0}` key in "
                "`pyqalxapi_list_response_dict` and for"
                " it to be a list".format(self._data_key)
            )
        # Cast all the entities in data to be an instance of `self.child`
        keyfile = kwargs.pop("keyfile", None)
        self[self._data_key] = [
            self.child(e, keyfile=keyfile)
            for e in pyqalxapi_list_response_dict[self._data_key]
        ]


class AggregationResult(QalxListEntity):
    """
    Class for aggregation result responses from the API
    """


class QalxEntity(ObjectDict):
    """Base class for qalx entities_response.

    QalxEntity children need to be populated with either a
    `requests.models.Response` which is the type returned by the methods
    on `pyqalxapi.api.PyQalxAPI` or with a `dict`.

    Entities can behave either like a dict or attribute lookups can be used
    as getters/setters

    >>> class AnEntity(QalxEntity):
    ...     pass
    >>> c = AnEntity({"guid":"123456789", "info":{"some":"info"}})
    >>> # dict style lookups
    >>> c['guid']
    '123456789'
    >>> # attribute style lookups
    >>> c.guid
    '123456789'


    :param pyqalxapi_dict: a 'dict' representing a qalx entity object to
        populate the entity
    :type pyqalxapi_dict: dict
    """

    entity_type: str

    def __init__(self, pyqalxapi_dict, *args, **kwargs):
        super().__init__(pyqalxapi_dict)

    @classmethod
    def _chunks(cls, _iterable, chunk_size, fillvalue=None):
        """
        Collect data into fixed-length chunks or blocks"
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        Taken from the itertools documentation
        """
        args = [iter(_iterable)] * chunk_size
        return zip_longest(fillvalue=fillvalue, *args)

    @classmethod
    def pluralise(cls):
        """
        Pluralises the entity type
        """
        vowels = {"a", "e", "i", "o", "u", "y"}
        singular = cls.entity_type
        plural = singular + "s"
        if len(singular) > 1 and singular.endswith("y"):
            if singular[-2] not in vowels:
                plural = singular[:-1] + "ies"
        return plural

    def __super_setattr__(self, name, value):
        """
        Convenience method for setting proper attributes on the entity class.
        Because we are using `ObjectDict` if we did
        `self.<name> = value` this would set the
        dict key of `<name>` which we don't want.  So we call
        the supermethod to properly set the attribute
        :param name: The name of the attribute to set
        :param value: The value of the attribute to set
        """
        super(ObjectDict, self).__setattr__(name, value)

    def __dir__(self):
        """
        By default `ObjectDict` __dir__ only returns the keys on the dict.
        We want it to return everything as normal as entities might have
        methods that the user needs to know about
        """
        return super(ObjectDict, self).__dir__()


class QalxFileEntity(QalxEntity):
    _file_bytes = None
    _file_key = "file"

    def __init__(self, *args, **kwargs):
        """
        QalxFileEntity subclasses can be provided with a `keyfile` kwarg.
        This gets set on the instance of the entity and will allow for seamless
        decryption of files
        """
        self.__super_setattr__("_key_file", kwargs.pop("keyfile", None))
        super(QalxFileEntity, self).__init__(*args, **kwargs)

    def read_file(self):
        """
        If this Item contains a file, will read the file data and cache it
        against the Item.

        :param key_file:    The path to the encryption key to be used for
                            decrypting the file data. If it is not provided
                            and the data was encrypted, the data will be read
                            in the encrypted format
        :return:    The content of the URL as a bytes object.  Accessible from
                    the `_file_bytes` attribute
        :raises: pyqalx.errors.QalxError
        """
        if not self.get(self._file_key):
            raise QalxError("Item doesn't have file data.")
        else:
            response = requests.get(url=self[self._file_key]["url"])
            if response.ok:
                file_bytes = response.content
                if (
                    self[self._file_key].get("keyfile") is not None
                    and self._key_file is not None
                ):
                    qalx_encrypt = QalxEncrypt(self._key_file)
                    file_bytes = qalx_encrypt.decrypt(file_bytes)
                self.__super_setattr__("_file_bytes", file_bytes)
                return self._file_bytes
            else:
                raise QalxError(
                    "Error with file retrieval: \n\n" + response.text
                )

    def save_file_to_disk(self, filepath, filename=None):
        """
        If this Item contains a file, will read the file from the URL (or from
        the cached bytes on the instance) and save the file to disk.  Provide
        an optional `filename` argument if you don't want to use the same
        filename as the one stored on the Item

        :param filepath: The path where this file should be saved
        :type filepath: str
        :param filename: The optional name of this file. Defaults to the name
            of the file on the instance
        :type filename: str
        :param key_file:    The path to the encryption key to be used for
                            decrypting the file data. If it is not provided
                            and the data was encrypted, the data will be read
                            in the encrypted format
        :type key_file:     str
        :raises: pyqalx.errors.QalxError
        """
        if filename is None:
            filename = self[self._file_key]["name"]
        if self._file_bytes is None:
            self.read_file()
        _filepath = os.path.join(filepath, filename)
        with open(_filepath, "wb") as f:
            f.write(self._file_bytes)
        return _filepath


class QalxQueueableEntity(QalxEntity):
    """
    A mixin which allows an entity to be sumitted to queue
    """

    @classmethod
    def add_to_queue(cls, payload, queue, children=False, **message_kwargs):
        """
        Submits an entity, entity guid or list of entities/entity guids
        to the given queue
        :param payload:subclassed instance of ~entities.entity.QalxEntity,
        a guid, or a list containing a combination of both
        :param queue: An instance of ~entities.queue.Queue
        Usage:
        :type queue:~entities.queue.Queue
        :param children:Whether we are submitting the child entities to the
        queue rather than the given entity
        :type children: bool
        Example usage for an Item
        Item.add_to_queue(item, queue)
        Item.add_to_queue(item.guid, queue)
        Item.add_to_queue([item, item.guid], queue)
        """
        entity_type = cls.entity_type
        if children:
            #  Get the children off this entity and submit them to the queue
            if not hasattr(cls, "child_entity_class"):
                raise QalxQueueError(
                    f"`{cls}` is not an entity that " f"has children."
                )
            if not isinstance(payload, cls):
                raise QalxQueueError(
                    f"Expected payload of type `{cls}"
                    f"` when calling `{cls.__name__}.add_to_queue` "
                    f"with `children=True`. "
                    f"Got `{type(payload)}`."
                )
            # SQS needs the guid as a string - not UUID
            message_kwargs[f"parent_{entity_type}_guid"] = str(payload["guid"])
            entity_type = cls.child_entity_class.entity_type
            payload = list(payload[cls.child_entity_class.pluralise()].values())
        return queue._send_message(
            payload=payload, entity_type=entity_type, **message_kwargs
        )
